{ **************************
  TOxYDSocket Class - by Ueli Kistler, 2002
  Based on COxYDSocket (Grgory Wilmes, www.engagesecurity.com) and ported to Delphi
  **************************
}

//$DEFINE DEBUG
unit OxYDSocket;

interface

uses Windows, Messages, SysUtils, StrUtils, Registry;

const   BROWSER           = ' - Microsoft Internet Explorer';
        BROWSER_EXE       = 'iexplore.exe';

        szTitleSize       = 256; // read max 256 caracters from Windows title
        szClassNameSize   = 1024-1;

type
        Stat = record      // Window handles
	        hWnd: HWND;
        end;

var
        // Windows informations
        szTitle     : Array[0..szTitleSize] of Char; // Window title
        winds       : Array[0..999] of Stat;       // Window handles
        nCount      : Integer;                     // Array count

        m_hWndTmp   : HWND;                        // Temporary value for a handle
        hProcess    : HWND;                     // hProcess Handle value

type

TOxyDError = (
        ERROR__UNKNOWN_ERROR,
        ERROR__UNABLE_TO_QUERY_VALUE_FROM_REGISTRY,
        ERROR__UNABLE_TO_FIND_BROWSER_RUNNING_DIRECTORY,
        ERROR__WRONG_DEFAULT_BROWSER,
        ERROR__UNABLE_TO_CREATE_A_NEW_PROCESS,
        ERROR__FAILED_TO_ENUMERATE_ALL_TOPLEVEL_WINDOWS,
        ERROR__FAILED_TO_FIND_BROWSER_WINDOW,
        ERROR__UNABLE_TO_TERMINATE_THE_PROCESS,
        ERROR__THE_BROWSER_IS_NOT_RUNNING
        );

TOxYDSocket = class(TObject)
        private
         m_hWnd, m_hWndUrl       : HWND;         // Windows handles
         // m_pWnd, m_pWndUrl    Pointer to Window
	 m_Rect                  : TRect;        // TRect to set window position

	 m_nLastError            : TOxYDError;      // Last error value
	 m_bBrowserRunning       : Boolean;      // Browser running flag

	 m_strRemoteAddress, m_strRemotePort    : String; // Remotehost, Port
         m_strWindowOldText                     : PChar; // Window title

        protected

 	public
		constructor Create;
		destructor Destroy; override;

                // Public functions

                // Browser running?
		function IsBrowserRunning:Boolean;

                // Connection handlers
                function Connect(strRemoteAddress: String; nRemotePort: Word):Boolean; overload; stdcall;
		function Connect(strRemoteAddress, strRemotePort: String):Boolean; overload; stdcall;

                // Shutdown
		function ShutDown:Boolean; // ShutDown socket, disconnect

                // Sending handler
		function SendRaw(strTxt: String):Boolean;       // Send a text
	        function ReceiveRaw(var strTxt:String):Boolean; // Receive text and write to strText

		function GetRemoteAddress:String;       // Result = Remote host
		function GetRemotePort:String;          // Result = Remote port

                function GetLastError:Integer;          // Get last error
	        function FormatMessage(nLastError: Integer):String; // Format Error message}
        end;

       // Define local
        function GetWindowProcX(hWnd: HWND; lParam: DWORD):Boolean; stdcall;
        function GetWindowProc(hWnd: HWND; lParam: DWORD):Boolean; stdcall;
        function GetChildWindow(hWnd: HWND; strCmp: String):HWND; stdcall;

 implementation

function GetWindowProcX(hWnd: HWND; lParam: DWORD):Boolean; export; stdcall;
var
        dwID            : DWORD;
begin
 GetWindowThreadProcessId(hWnd, @dwID); // Get Thread ID into dwID
 if (dwID=lParam) then begin
  GetWindowText(hWnd, szTitle, szTitleSize);
  if (RightStr(szTitle,length(BROWSER)) = BROWSER) then begin
    m_hWndTmp := hWnd;
    result:= True;
    exit;
  end;
 end;
 Result := True; // We need to check all windows
end;

function GetWindowProc(hWnd: HWND; lParam: DWORD):Boolean; export; stdcall;
begin
 GetWindowText(hWnd, szTitle, szTitleSize);

 winds[nCount].hWnd := hWnd;
 Inc(nCount);

 result:=true;
end;

function GetChildWindow(hWnd: HWND; strCmp: String):HWND; export; stdcall;
var
        szClassName     : array[0..szClassNameSize] of Char; // Window title
        i               : Cardinal;
begin
 nCount         := 0;
 if not (EnumChildWindows(hWnd, @GetWindowProc, 0)) then begin
    // Failed
    result      := 0;
    exit;
 end;

 while (EnumChildWindows(winds[nCount-1].hWnd, @GetWindowProc, 0)) do;

 for i:=0 to nCount-1 do begin
  if (GetClassName(winds[i].hWnd, szClassName, szClassNameSize)>0) then begin
     if (szClassName=strCmp) then begin
      result:=winds[i].hWnd;
      exit;
     end;
  end;
 end;

 result:=0;
end;


{
/* The constructor. Initializes all variables to an inital value. */
/*
}
constructor TOxYDSocket.Create;
begin
 m_hWnd         := 0;
 m_hWndUrl      := 0;
 // m_pWnd := 0
 // m_pWndUrl := 0

 m_nLastError          := ERROR__UNKNOWN_ERROR;
 m_bBrowserRunning     := False;

 m_strRemoteAddress    := '';
 m_strRemotePort       := '';

 // Allocate some memory
 m_strWindowOldText    := '';

 szTitle     := '';
 nCount      := 0;

 m_hWndTmp   := 0;
 hProcess    := 0;
end;

destructor TOxYDSocket.Destroy;
begin
 ShutDown;
 inherited Destroy;
end;

function TOxYDSocket.IsBrowserRunning:Boolean;
begin
 result := m_bBrowserRunning;
end;

function TOxYDSocket.Connect(strRemoteAddress: String; nRemotePort: Word):Boolean;
var strRemotePort: String;
begin
 strRemotePort := IntToStr(nRemotePort);
 result:=Connect(strRemoteAddress, strRemotePort);
end;

function TOxYDSocket.Connect(strRemoteAddress, strRemotePort: String):Boolean;
var
        szBuffer        : String;
        szBufferCLSID   : String;

        nPos            : Integer;      // =0

        strRunningDir   : String;

        stInfo          : STARTUPINFO;
        prInfo          : PROCESS_INFORMATION;
        bResult         : Boolean;      // CreateProcess result

        // Delphi implemenation of OxYDSocket accesses the registry using TRegistry class
        RegQuery        : TRegistry;

        // TRect to set a window position
        Rect            : TRect;

        // EnumWindows
        i               : Integer;
begin
 // Initialize
 szBuffer       := '';
 szBufferCLSID  := '';
 nPos           := 0;
 nCount         := 0;

 // Create a hidden browser window now
 if (m_HWnd = 0) then begin
     RegQuery           := TRegistry.Create;
     RegQuery.RootKey   := HKEY_CLASSES_ROOT;

     // Get the Explorer name
     with RegQuery do begin

      if OpenKey('htmlfile\CLSID',false) then begin
         szBufferCLSID   := ReadString('');  // Get CLSID

         if szBufferCLSID='' then begin
           m_nLastError  := ERROR__UNABLE_TO_QUERY_VALUE_FROM_REGISTRY;
           result        := False;
           exit;
         end;
         CloseKey;
      end
      else begin
       m_nLastError  := ERROR__UNABLE_TO_QUERY_VALUE_FROM_REGISTRY;
       result := False;
       exit;
      end;

      szBuffer   := format('CLSID\%s\DefaultIcon', [szBufferCLSID]);
      if OpenKey(szBuffer,false) then begin
         szBuffer   := ReadString('');  // Get CLSID

         if szBuffer='' then begin
           m_nLastError  := ERROR__UNABLE_TO_QUERY_VALUE_FROM_REGISTRY;
           result        := False;
           exit;
         end;
         CloseKey;
      end
      else begin
       m_nLastError  := ERROR__UNABLE_TO_QUERY_VALUE_FROM_REGISTRY;
       result := False;
       exit;
      end;

      strRunningDir     := szBuffer;            // Get Browser directory\file, number

      nPos      := Pos(',',strRunningDir);
      if nPos = 0 then begin
        m_nLastError  := ERROR__UNABLE_TO_FIND_BROWSER_RUNNING_DIRECTORY;
        result        := False;
        exit;
      end;

      strRunningDir     := LeftStr(strRunningDir,nPos-1); // Extract directory\file

      // Is Microsoft Internet Explorer your default Browser?
      //
      if not (RightStr(strRunningDir,length(BROWSER_EXE)) = BROWSER_EXE) then begin
        m_nLastError  := ERROR__WRONG_DEFAULT_BROWSER;
        result        := False;
        exit;
      end;

      // Free RegQuery
      regquery.Free;

     end; // With RegQuery. Done

     {
     /* Launch browser */
     }

     // Initialize stInfo
     FillChar(stInfo, SizeOf(TStartupInfo), 0);  // ZeroMemory
     stInfo.cb          := Sizeof(TStartupInfo);
     stInfo.dwFlags     := STARTF_USESHOWWINDOW;


    {$IFNDEF DEBUG}
     stInfo.wShowWindow := SW_HIDE;
    {$ENDIF}

     bResult    := CreateProcess(nil,
                                 Pchar(strRunningDir),
                                 nil,
                                 nil,
                                 True,
                                 CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
                                 nil,
                                 nil,
                                 stInfo,
                                 prInfo);

     if not (bResult) then begin
        m_nLastError    := ERROR__UNABLE_TO_CREATE_A_NEW_PROCESS;
        result          := False;
        exit;
     end;

     hProcess   := prInfo.hProcess;

     // What handle does our Internet Explorer !window! have?

     m_hWndTmp  := 0;

     while (m_hWndTmp=0) do begin
        if not (EnumWindows(@GetWindowProcX, prInfo.dwProcessId)) then begin
           m_nLastError    := ERROR__FAILED_TO_FIND_BROWSER_WINDOW;
           result          := False;
           exit;
        end;

        if (m_hWndTmp = 0) then begin

          i:=nCount+1;
          if (i > 200) then begin     // 50ms * 200 = 10s
              TerminateProcess(hProcess, 0);

              m_nLastError        := ERROR__FAILED_TO_FIND_BROWSER_WINDOW;
              result              := False;
              exit;
          end;

          Sleep(50);
        end;

     end; // While (m_hWndTmp = 0)

    m_hWnd     := m_hWndTmp;

     if m_hWnd<>0 then begin      // Got a valid window handle?

        m_strRemoteAddress      := strRemoteAddress;
        m_strRemotePort         := strRemotePort;

        // Get Browser window handle
        // This was needed in Visual C++, but in Delphi we do this different
        // m_pWnd = CWnd::FromHandle(m_hWnd);

        // Get Windows positions
        GetWindowRect(m_hWnd,Rect);

        // Copy the informations to global variable
        m_Rect := Rect;

        Rect.Left       := Rect.Left-5000;  // Rect.Left -= 5000
        Rect.Right      := Rect.Left-5000;  //     .Right

        // Move the window
        MoveWindow(m_hWnd, Rect.Left, Rect.Top,m_Rect.Right-m_Rect.Left, m_Rect.Bottom-m_Rect.Top,False);

        // Get window title
        GetMem(m_strWindowOldText,GetWindowTextLength(m_hWnd)+1);
        GetWindowText(m_hWnd, m_strWindowOldText,GetWindowTextLength(m_hWnd)+1); // Get whole window title

        // Get URL editbox handle
        m_hWndUrl       :=  GetChildWindow(m_hWnd, 'Edit');
        if m_hWndUrl<>0 then begin
           // Set tool window style
           
           // ... hide from taskbar
           SetWindowLong(m_hWnd, GWL_EXSTYLE, WS_EX_TOOLWINDOW);

           // Set Flag
           m_bBrowserRunning    := True;
        end else begin
           m_nLastError    := ERROR__FAILED_TO_ENUMERATE_ALL_TOPLEVEL_WINDOWS;
           result          := False;
           exit;
        end;

     end else // if not (m_hWnd =0)
      m_bBrowserRunning := False;

 end; // if (m_HWnd = 0)
 result := m_bBrowserRunning;
end;

function TOxYDSocket.Shutdown:Boolean;
begin
 if (IsBrowserRunning) then begin
    // Hide Browser window
    {$IFNDEF DEBUG}
    ShowWindow(m_hWnd, SW_HIDE);
    {$ENDIF}


    MoveWindow(m_hWnd, m_Rect.left, m_Rect.Top, m_Rect.Right-m_Rect.Left, m_Rect.Bottom-m_Rect.Top, True);

    if not (TerminateProcess(hProcess, 0)) then begin
       m_nLastError := ERROR__UNABLE_TO_TERMINATE_THE_PROCESS;
       result := False;
       exit;
    end;

    m_hWnd            := 0;
    m_hWndTmp         := 0;

    m_bBrowserRunning := False;

    Result := True;
 end else Result := False;
end;

function TOxYDSocket.SendRaw(strTxt: String):Boolean;
var
        urltext : String;
begin
 if IsBrowserRunning then begin
    // Set URL text
    if (m_strRemotePort='') or (m_strRemoteport='80') then
       urltext := 'http://'+ m_strRemoteAddress+'/'+strTxt
    else
       urltext := 'http://'+ m_strRemoteAddress+':'+m_strRemotePort+'/'+strTxt;

    SendMessage(m_hWndUrl, WM_SETTEXT, 0, Integer(PChar(urltext)));

    // Simulate pressing the ENTER key
    SendMessage(m_hWndUrl, WM_KEYDOWN, 13,0); // Use Sendmessage to prevent return before action is done
    SendMessage(m_hWndUrl, WM_CHAR, 13,0);
    SendMessage(m_hWndUrl, WM_KEYUP, 13,0);

    result := True;
    exit;
 end;

 m_nLastError := ERROR__THE_BROWSER_IS_NOT_RUNNING;
 result := False;
end;

function TOxYDSocket.ReceiveRaw(var strTxt: String):Boolean;
const nLen = Length(BROWSER);
var
        strWindowText: PChar; // Window title
begin

 if IsBrowserRunning then begin
    GetMem(strWindowText, GetWindowTextLength(m_hWnd)+1);
    repeat
        Sleep(50);
        GetWindowText(m_hWnd, strWindowText, GetWindowTextLength(m_hWnd)+1);
    until not (PChar(strWindowText) = m_strWindowOldText);

    if (RightStr(String(strWindowText), nLen) = BROWSER) then
        strWindowText := PChar(LeftStr(String(strWindowText), length(strWindowText)-nLen))
    else begin
      m_nLastError := ERROR__THE_BROWSER_IS_NOT_RUNNING;
      result := False;
      exit;
    end;

    m_strWindowOldText  := PChar(strWindowText);
    strTxt              := String(strWindowText);

    result := True;
 end;

 m_nLastError := ERROR__THE_BROWSER_IS_NOT_RUNNING;
 result := False;
end;

function TOxYDSocket.GetRemoteAddress:String;
begin
 result := m_strRemoteAddress;
end;

function TOxYDSocket.GetRemotePort:String;
begin
 result := m_strRemotePort;
end;

function TOxYDSocket.GetLastError:Integer;
begin
 result := Integer(m_nLastError);
end;

// Format error message
function TOxYDSocket.FormatMessage(nLastError: Integer):String;
var
        strLastError: String;
begin
 case nLastError of
  Integer(ERROR__UNKNOWN_ERROR): strLastError := 'Unknown error';
  Integer(ERROR__UNABLE_TO_QUERY_VALUE_FROM_REGISTRY): strLastError := 'Unable to query value from registry';
  Integer(ERROR__UNABLE_TO_FIND_BROWSER_RUNNING_DIRECTORY): strLastError := 'Unable to find browser directory';
  Integer(ERROR__WRONG_DEFAULT_BROWSER): strLastError := 'Wrong default browser';
  Integer(ERROR__UNABLE_TO_CREATE_A_NEW_PROCESS): strLastError := 'Unable to create a new process';
  Integer(ERROR__FAILED_TO_ENUMERATE_ALL_TOPLEVEL_WINDOWS): strLastError := 'Failed to enumerate all top-level windows';
  Integer(ERROR__FAILED_TO_FIND_BROWSER_WINDOW): strLastError := 'Failed to find browser window';
  Integer(ERROR__UNABLE_TO_TERMINATE_THE_PROCESS): strLastError := 'Unable to terminate the process';
  Integer(ERROR__THE_BROWSER_IS_NOT_RUNNING): strLastError := 'The browser is not running';
 else
  strLastError := 'Unknown error code';
 end;

 result := strLastError;

end;

end.
